#include "bin_tree.h"
#include <assert.h>
#include <math.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>

static node *position(arbre tree, cle c) {
    node *crt = tree;
    if (NULL != crt) {
        while (c != crt->key && NULL != crt->child[(c > crt->key)]) {
            crt = crt->child[(c > crt->key)];
        }
    }
    return crt;
}

bool arbre_is_empty(arbre tree) {
    return NULL == tree;
}

int arbre_depth(arbre tree) {
    if (arbre_is_empty(tree)) {
        return 0;
    } else {
        return 1 +
               fmax(arbre_depth(tree->child[0]), arbre_depth(tree->child[1]));
    }
}

int arbre_size(arbre tree) {
    if (arbre_is_empty(tree)) {
        return 0;
    } else {
        return 1 + arbre_size(tree->child[0]) + arbre_size(tree->child[1]);
    }
}

static node *create_node(int val) {
    node *nd = calloc(1, sizeof(node));
    nd->key  = val;
    return nd;
}

arbre arbre_insert(arbre tree, cle c) {
    if (arbre_is_empty(tree)) {
        tree = create_node(c);
    } else {
        node *nd                 = position(tree, c);
        nd->child[(c > nd->key)] = create_node(c);
    }
    return tree;
}

static node *parent(arbre tree, node *nd) {
    assert(!arbre_is_empty(tree) && NULL != nd);
    node *parent = NULL;
    if (nd != tree) {
        node *crt = tree;
        cle c     = nd->key;
        do {
            parent = crt;
            if (c != crt->key) {
                crt = crt->child[(c > crt->key)];
            }
        } while (crt != nd);
    }
    return parent;
}

bool arbre_search(arbre tree, cle c) {
    node *nd = position(tree, c);
    return (NULL != nd && c == nd->key);
}

bool arbre_delete(arbre *tree, cle c) {
    node *nd = position(*tree, c);
    if (NULL == nd || c != nd->key) {
        return false;
    }
    // noeud terminal ou avec 1 enfant
    if (arbre_is_empty(nd->child[0]) || arbre_is_empty(nd->child[1])) {
        node *nd_parent = parent(*tree, nd);
        if (NULL == nd_parent) { // nd est la racine
            *tree = nd->child[(NULL != nd->child[1])];
        } else {
            nd_parent->child[(nd == nd_parent->child[1])] =
                nd->child[(NULL != nd->child[1])];
        }
        free(nd);
    } else {                                    // noeud interne (2 enfants)
        node *next = position(nd->child[1], c); // next a 0 ou 1 enfant
        cle tmp    = next->key;
        bool res   = arbre_delete(tree, tmp);
        nd->key    = tmp;
    }
    return true;
}

bool arbre_delete_bis(arbre *tree, cle c) {
    node *nd = position(*tree, c);

    if (NULL == nd || c != nd->key) {
        return false;
    }

    // terminal node
    if (arbre_is_empty(nd->child[0]) && arbre_is_empty(nd->child[1])) {
        node *nd_parent = parent(*tree, nd);
        if (NULL == nd_parent) { // single node tree
            *tree = NULL;
        } else {
            nd_parent->child[(nd == nd_parent->child[1])] = NULL;
        }
        free(nd);
        return true;
    }

    for (int ind = 0; ind < 2; ind++) {
        if (!arbre_is_empty(nd->child[ind])) {
            node *next = position(nd->child[ind], c);
            int val    = next->key;
            if (NULL == nd->child[ind]->child[ind ^ 1]) {
                nd->child[ind] = next->child[ind];
                free(next);
            } else {
                bool res = arbre_delete(tree, next->key);
            }
            nd->key = val;
            return true;
        }
    }
    return false;
}

void arbre_print(arbre tree, int N) {
    if (N <= 0) {
        N = 1;
    }
    if (NULL != tree) {
        arbre_print(tree->child[1], N + 1);
        for (int i = 0; i < N; i++) {
            printf("    ");
        }
        printf("%d\n", tree->key);
        arbre_print(tree->child[0], N + 1);
    }
}
